#include <math.h>
#include <pthread.h>
#include <open62541/server.h>
#include "dronedge_protocol.h"
#include "pubsub.h"
#ifdef WITH_AUTOPILOT
#include "autopilot_API.h"
#endif

#define DRONE_VARIABLE_NB 4
#define SUBSCRIBER_VARIABLE_NB 1

UA_Int64 positionArray[POSITION_ARRAY_SIZE] = { 0 };
UA_UInt32 positionArrayDims[] = {POSITION_ARRAY_SIZE};

UA_Double speedArray[SPEED_ARRAY_SIZE] = { 0 };
UA_UInt32 speedArrayDims[] = {SPEED_ARRAY_SIZE};

UA_String initial_ua_string = {
    .length = 0,
    .data = NULL,
};

VariableData droneVariableArray[] = {
    {
        .name = "message",
        .description = "Message to send to the other drones",
        .value = &initial_ua_string,
        .type = UA_TYPES_STRING,
        .builtInType = UA_NS0ID_STRING,
        .valueRank = UA_VALUERANK_SCALAR,
        .arrayDimensionsSize = 0,
        .arrayDimensions = NULL,
        .getter.getString = get_message,
    },
    {
        .name = "positionArray",
        .typeName = "Position Array Type",
        .description = "Position Array",
        .value = &positionArray,
        .type = UA_TYPES_INT64,
        .builtInType = UA_NS0ID_INT64,
        .valueRank = UA_VALUERANK_ONE_DIMENSION,
        .arrayDimensionsSize = 1,
        .arrayDimensions = positionArrayDims,
#ifdef WITH_AUTOPILOT
        .getter.getArray = (void * (*)(void))getPositionArray,
#endif
    },
    {
        .name = "speedArray",
        .typeName = "Speed Array Type",
        .description = "Speed Array",
        .value = &speedArray,
        .type = UA_TYPES_FLOAT,
        .builtInType = UA_NS0ID_FLOAT,
        .valueRank = UA_VALUERANK_ONE_DIMENSION,
        .arrayDimensionsSize = 1,
        .arrayDimensions = speedArrayDims,
#ifdef WITH_AUTOPILOT
        .getter.getArray = (void * (*)(void))getSpeedArray,
#endif
    },
    {
        .name = "log",
        .description = "Message to send to the other drones",
        .value = &initial_ua_string,
        .type = UA_TYPES_STRING,
        .builtInType = UA_NS0ID_STRING,
        .valueRank = UA_VALUERANK_SCALAR,
        .arrayDimensionsSize = 0,
        .arrayDimensions = NULL,
        .getter.getString = get_log,
    },
};

VariableStruct droneVariables = {
    .nbVariable = DRONE_VARIABLE_NB,
    .variableArray = droneVariableArray,
};

VariableData subscriberVariableArray[] = {
    {
        .name = "message",
        .description = "Message to send to the other drones",
        .value = &initial_ua_string,
        .type = UA_TYPES_STRING,
        .builtInType = UA_NS0ID_STRING,
        .valueRank = UA_VALUERANK_SCALAR,
        .arrayDimensionsSize = 0,
        .arrayDimensions = NULL,
        .getter.getString = get_message,
    },
};

VariableStruct subscriberVariables = {
    .nbVariable = SUBSCRIBER_VARIABLE_NB,
    .variableArray = subscriberVariableArray,
};

static JSClassID jsDroneClassId;
static JSClassID jsPositionClassId;

static UA_Boolean pubsubShouldRun = true;
static UA_Boolean pubsubExited = true;

static UA_UInt32 nbDrone;
static UA_UInt32 nbSubscriber;
static JSValueConst *droneObjectIdList;
static StrQueue sendMessageQueue = {
    .head = NULL,
    .tail = NULL,
};
UA_String currentSendMessage;
static StrQueue sendLogQueue = {
    .head = NULL,
    .tail = NULL,
};
UA_String currentSendLog;

bool isADrone;

PubsubVariableIDs *pubsubIdsArray;
const UA_UInt32 VARIABLE_ID_BASE = 50000;

// Position class functions

static void js_position_finalizer(JSRuntime *rt, JSValue val)
{
    JSPositionData *s = (JSPositionData *) JS_GetOpaque(val, jsPositionClassId);
    js_free_rt(rt, s);
}

#ifdef WITH_AUTOPILOT
static JSValue js_new_position(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    JSPositionData *s;
    JSValue obj;
    UA_Int64 *newPositionArray;
    obj = JS_NewObjectClass(ctx, (int) jsPositionClassId);
    if (JS_IsException(obj))
        return obj;
    s = (JSPositionData *) js_mallocz(ctx, sizeof(*s));
    if (!s) {
        JS_FreeValue(ctx, obj);
        return JS_EXCEPTION;
    }
    newPositionArray = getPositionArray();
    s->latitude = (double)positionArray[LATITUDE] / 1e7;
    s->longitude = (double)positionArray[LONGITUDE] / 1e7;
    s->altitude = (UA_Double)(positionArray[REL_ALTITUDE] / 1000);
    s->timestamp = positionArray[TIMESTAMP];
    JS_SetOpaque(obj, s);
    free(newPositionArray);
    return obj;
}
#endif

static JSValue js_position_get(JSContext *ctx, JSValueConst thisVal, int magic)
{
    JSPositionData *s = (JSPositionData *) JS_GetOpaque2(ctx, thisVal, jsPositionClassId);
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewFloat64(ctx, s->latitude);
        case 1:
            return JS_NewFloat64(ctx, s->longitude);
        case 2:
            return JS_NewFloat64(ctx, s->altitude);
        case 3:
            return JS_NewInt64(ctx, s->timestamp);
        default:
            return JS_EXCEPTION;
    }
}

static JSClassDef jsPositionClass = {
    "Position",
    .finalizer = js_position_finalizer,
};

static const JSCFunctionListEntry js_position_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("latitude", js_position_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("longitude", js_position_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("altitude", js_position_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("timestamp", js_position_get, NULL, 3),
};

// Drone class functions

static void delete_str_node(struct strNode *node) {
    free(node->str);
    free(node);
}

static void cleanQueue(StrQueue *pQueue)
{
    struct strNode *current;

    while (pQueue->head != NULL) {
        current = pQueue->head;
        pQueue->head = current->next;
        delete_str_node(current);
    }
}

static void js_drone_finalizer(JSRuntime *rt, JSValue val)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(val, jsDroneClassId);
    cleanQueue(&(s->receiveMessageQueue));
    cleanQueue(&(s->receiveLogQueue));
    js_free_rt(rt, s);
}

static JSValue js_drone_ctor(JSContext *ctx, JSValueConst newTarget, 
                             int argc, JSValueConst *argv)
{
    JSDroneData *s;
    JSValue obj = JS_UNDEFINED;
    JSValue proto;
    uint32_t uint32;
    
    s = (JSDroneData *) js_mallocz(ctx, sizeof(*s));
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToUint32(ctx, &uint32, argv[0]))
        goto fail;
    s->id = (uint16_t)uint32;
    s->receiveMessageQueue = (StrQueue){
        .head = NULL,
        .tail = NULL,
    };
    s->receiveLogQueue = (StrQueue){
        .head = NULL,
        .tail = NULL,
    };

    proto = JS_GetPropertyStr(ctx, newTarget, "prototype");
    if (JS_IsException(proto))
        goto fail;
    obj = JS_NewObjectProtoClass(ctx, proto, jsDroneClassId);
    JS_FreeValue(ctx, proto);
    if (JS_IsException(obj))
        goto fail;
    JS_SetOpaque(obj, s);
    return obj;
 fail:
    js_free(ctx, s);
    JS_FreeValue(ctx, obj);
    return JS_EXCEPTION;
}

static JSValue js_drone_init(JSContext *ctx, JSValueConst thisVal,
                             int argc, JSValueConst *argv)
{
    int nb;
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &nb, argv[0]))
        return JS_EXCEPTION;
    droneObjectIdList[nb] = thisVal;
    return JS_UNDEFINED;
}

static JSValue readDroneDataStr(JSContext *ctx, StrQueue *pQueue, bool keepAtLeastAnElement)
{
    JSValue res;
    struct strNode *current;

    current = pQueue->head;
    if (current != NULL) {
        res = JS_NewString(ctx, current->str);
        if (current->next != NULL) {
            pQueue->head = current->next;
            delete_str_node(current);
        } else {
            if (!keepAtLeastAnElement) {
                pQueue->head = (pQueue->tail = NULL);
                delete_str_node(current);
            }
        }
    } else {
        res = JS_NewString(ctx, "");
    }
    return res;
}

static JSValue js_drone_get(JSContext *ctx, JSValueConst thisVal, int magic)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewUint32(ctx, s->id);
        case 1:
            return JS_NewFloat64(ctx, s->latitude);
        case 2:
            return JS_NewFloat64(ctx, s->longitude);
        case 3:
            return JS_NewFloat64(ctx, s->altitudeAbs);
        case 4:
            return JS_NewFloat64(ctx, s->altitudeRel);
        case 5:
            return JS_NewFloat64(ctx, s->yaw);
        case 6:
            return JS_NewFloat64(ctx, s->speed);
        case 7:
            return JS_NewFloat64(ctx, s->climbRate);
        case 8:
            return readDroneDataStr(ctx, &(s->receiveMessageQueue), true);
        case 9:
            return readDroneDataStr(ctx, &(s->receiveLogQueue), false);
        case 10:
            return JS_NewInt64(ctx, s->timestamp);
        default:
            return JS_EXCEPTION;
    }
}

static void addStrToQueue(const char *str, StrQueue *pQueue)
{
    struct strNode *newNode;
    newNode = (struct strNode*)malloc(sizeof(struct strNode));
    newNode->str = strdup(str);
    newNode->next = NULL;

    if (pQueue->tail == NULL) {
        pQueue->head = pQueue->tail = newNode;
    } else {
        pQueue->tail->next = newNode;
        pQueue->tail = newNode;
    }
}

static JSValue add_jsstr_to_queue(JSContext *ctx, JSValueConst jsStr,
                                  StrQueue *pQueue)
{
    const char *str;
    str = JS_ToCString(ctx, jsStr);
    if (strlen(str) > MAX_STR_SIZE) {
        UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "String too long");
        return JS_EXCEPTION;
    }

    addStrToQueue(str, pQueue);
    JS_FreeCString(ctx, str);
    return JS_UNDEFINED;
}

static JSValue js_drone_set_message(JSContext *ctx, JSValueConst thisVal,
                                    int argc, JSValueConst *argv)
{
    return add_jsstr_to_queue(ctx, argv[0], &sendMessageQueue);
}

static JSValue js_drone_set_log(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    return add_jsstr_to_queue(ctx, argv[0], &sendLogQueue);
}

static UA_Boolean UA_String_isEmpty(const UA_String *s) {
    return (s->length == 0 || s->data == NULL);
}

static void clear_str(UA_String *pstr) {
    if (!UA_String_isEmpty(pstr))
        UA_String_clear(pstr); 
}

static UA_String get_StrQueue_content(StrQueue *pQueue, UA_String currentStr)
{
    struct strNode *current;

    current = pQueue->head;
    if (current != NULL) {
        clear_str(&currentStr);
        currentStr = UA_STRING_ALLOC(current->str);
        pQueue->head = current->next == NULL ? (pQueue->tail = NULL) : current->next;
        delete_str_node(current);
    }
    return currentStr;
}

UA_String get_message(void)
{
    return get_StrQueue_content(&sendMessageQueue, currentSendMessage);
}

UA_String get_log(void)
{
    return get_StrQueue_content(&sendLogQueue, currentSendLog);
}

static JSClassDef jsDroneClass = {
    "Drone",
    .finalizer = js_drone_finalizer,
};

static const JSCFunctionListEntry js_drone_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("id", js_drone_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("latitude", js_drone_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("longitude", js_drone_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("altitudeAbs", js_drone_get, NULL, 3),
    JS_CGETSET_MAGIC_DEF("altitudeRel", js_drone_get, NULL, 4),
    JS_CGETSET_MAGIC_DEF("yaw", js_drone_get, NULL, 5),
    JS_CGETSET_MAGIC_DEF("speed", js_drone_get, NULL, 6),
    JS_CGETSET_MAGIC_DEF("climbRate", js_drone_get, NULL, 7),
    JS_CGETSET_MAGIC_DEF("message", js_drone_get, NULL, 8),
    JS_CGETSET_MAGIC_DEF("log", js_drone_get, NULL, 9),
    JS_CGETSET_MAGIC_DEF("timestamp", js_drone_get, NULL, 10),
    JS_CFUNC_DEF("init", 1, js_drone_init),
};

// pubsub functions

UA_UInt16 get_drone_id(UA_UInt32 nb) {
    JSDroneData *s =
        (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    return s->id;
}

VariableData pubsub_get_value(UA_String identifier) {
    VariableData varDetails;
    UA_DataType type;
    UA_Double *array;

    for(UA_UInt32 i = 0; i < droneVariables.nbVariable; i++) {
        UA_String name = UA_STRING(droneVariables.variableArray[i].name);
        if(UA_String_equal(&identifier, &name)) {
            varDetails = droneVariables.variableArray[i];
            switch(varDetails.valueRank) {
                case UA_VALUERANK_SCALAR:
                    switch(varDetails.type) {
                        case UA_TYPES_STRING:
                            *(UA_String*)varDetails.value =
                                varDetails.getter.getString();
                            break;
                        default:
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
                                         "UA_TYPE not handled");
                            break;
                    }
                    break;

                case UA_VALUERANK_ONE_DIMENSION:
                    type = UA_TYPES[varDetails.type];
                    size_t size = varDetails.arrayDimensions[0];
                    array = (UA_Double *) varDetails.getter.getArray();

                    if(type.pointerFree) {
                        memcpy(varDetails.value, array, type.memSize * size);
                    } else {
                        uintptr_t ptrs = (uintptr_t)array;
                        uintptr_t ptrd = (uintptr_t)varDetails.value;
                        UA_StatusCode retval = UA_STATUSCODE_GOOD;
                        for(size_t j = 0; j < size; ++j) {
                            retval |= UA_copy((void*)ptrs, (void*)ptrd, &type);
                            ptrs += type.memSize;
                            ptrd += type.memSize;
                        }
                        if(retval != UA_STATUSCODE_GOOD)
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
                                         "Failed to update array");
                    }

                    free(array);
                    break;

                default:
                    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
                                 "UA_VALUERANK not handled");
                    break;
            }
        }
    }

    return varDetails;
}

static void setDroneDataStr(void *data, StrQueue* pQueue)
{
    UA_String uaStr = *(UA_String*) data;
    char str[MAX_STR_SIZE];
    memcpy(str, uaStr.data, uaStr.length);
    str[uaStr.length] = '\0';
    addStrToQueue(str, pQueue);
}

static void pubsub_update_variables(UA_UInt32 id, const UA_DataValue *var, bool print)
{
    JSDroneData* s;
    UA_Int64* updatedPositionArray;
    UA_Float* updatedSpeedArray;

    for(UA_UInt32 i = 0; i < nbDrone + nbSubscriber; i++) {
        s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[i], jsDroneClassId);
        if (pubsubIdsArray[i].positionArray == id) {
            updatedPositionArray = (UA_Int64*) var->value.data;
            s->latitude = (double)updatedPositionArray[LATITUDE] / 1e7;
            s->longitude = (double)updatedPositionArray[LONGITUDE] / 1e7;
            s->altitudeAbs = (UA_Double)(updatedPositionArray[ABS_ALTITUDE] / 1000);
            s->altitudeRel = (UA_Double)(updatedPositionArray[REL_ALTITUDE] / 1000);
            s->timestamp = updatedPositionArray[TIMESTAMP];

            if (print) {
                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT,
                  "Received position of drone %d: %.6f ° %.6f ° %.2f m %.2f m",
                  s->id, s->latitude, s->longitude, s->altitudeAbs, s->altitudeRel);
            }
            return;
        } else if (pubsubIdsArray[i].speedArray == id) {
            updatedSpeedArray = (UA_Float*) var->value.data;
            s->yaw = updatedSpeedArray[YAW];
            s->speed = updatedSpeedArray[SPEED];
            s->climbRate = updatedSpeedArray[CLIMB_RATE];

            if (print) {
                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT,
                  "Received speed of drone %d: %.2f ° %.2f m/s %.2f m/s",
                  s->id, s->yaw, s->speed, s->climbRate);
            }
            return;
        } else if (pubsubIdsArray[i].message == id) {
            setDroneDataStr(var->value.data, &(s->receiveMessageQueue));
            return;
        } else if (pubsubIdsArray[i].log == id) {
            if (!isADrone) {
                setDroneDataStr(var->value.data, &(s->receiveLogQueue));
            }
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT,
                 "NodeId %i not found", id);
}

/*
 * arg 0 (string): IPv6 of the multicast group
 * arg 1 (string): port used for multicast communication
 * arg 2 (string): network interface used for multicast communication
 * arg 3 (int): ID of the drone
 * arg 4 (double): publication/subscription interval in ms
 * arg 5 (bool): true if the peer is a drone
*/
static JSValue js_run_pubsub(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv) {
    const char *ipv6;
    const char *port;
    const char *netIface;
    char *notConstNetIface;
    char urlBuffer[44];
    UA_UInt32 id;
    UA_Duration interval;
    VariableStruct variables;
    VariableStruct *pubsubVariableArray;
    UA_UInt32 nbPeer = nbDrone + nbSubscriber;
    int res;

    ipv6 = JS_ToCString(ctx, argv[0]);
    port = JS_ToCString(ctx, argv[1]);
    netIface = JS_ToCString(ctx, argv[2]);
    UA_snprintf(urlBuffer, sizeof(urlBuffer), "opc.udp://[%s]:%s/", ipv6, port);

    UA_String transportProfile =
        UA_STRING("http://opcfoundation.org/UA-Profile/Transport/pubsub-udp-uadp");

    UA_NetworkAddressUrlDataType networkAddressUrl;
    notConstNetIface = strdup(netIface);
    networkAddressUrl.networkInterface = UA_STRING(notConstNetIface);
    networkAddressUrl.url              = UA_STRING(urlBuffer);

    if (JS_ToUint32(ctx, &id, argv[3]))
        return JS_EXCEPTION;

    if (JS_ToFloat64(ctx, &interval, argv[4]))
        return JS_EXCEPTION;

    isADrone = JS_ToBool(ctx, argv[5]);
    variables = isADrone ? droneVariables : subscriberVariables;

    pubsubVariableArray = (VariableStruct *) malloc((nbPeer) * sizeof(VariableStruct));
    pubsubIdsArray = (PubsubVariableIDs *) malloc((nbPeer) * sizeof(PubsubVariableIDs));
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
      pubsubVariableArray[i] = droneVariables;
      PubsubVariableIDs ids = {
          .message = (UA_UInt32) DRONE_VARIABLE_NB * i + VARIABLE_ID_BASE,
          .positionArray = (UA_UInt32) DRONE_VARIABLE_NB * i + VARIABLE_ID_BASE + 1,
          .speedArray = (UA_UInt32) DRONE_VARIABLE_NB * i + VARIABLE_ID_BASE + 2,
          .log = (UA_UInt32) DRONE_VARIABLE_NB * i + VARIABLE_ID_BASE + 3,
      };
      pubsubIdsArray[i] = ids;
    }
    for(UA_UInt32 i = nbDrone; i < nbPeer; i++) {
      pubsubVariableArray[i] = subscriberVariables;
      PubsubVariableIDs ids = {
          .message = (UA_UInt32) DRONE_VARIABLE_NB * i + VARIABLE_ID_BASE,
      };
      pubsubIdsArray[i] = ids;
    }

    pubsubExited = false;
    res = runPubsub(UA_Log_Stdout, &transportProfile, &networkAddressUrl,
                    variables, id, pubsubVariableArray, pubsubIdsArray, nbPeer,
                    interval, get_drone_id, pubsub_get_value,
                    pubsub_update_variables, &pubsubShouldRun);
    pubsubExited = true;

    free(pubsubVariableArray);
    free(pubsubIdsArray);
    JS_FreeCString(ctx, ipv6);
    JS_FreeCString(ctx, port);
    free(notConstNetIface);
    JS_FreeCString(ctx, netIface);
    return JS_NewInt32(ctx, res);
}

static JSValue js_init_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    if (JS_ToUint32(ctx, &nbDrone, argv[0]))
        return JS_EXCEPTION;

    if (JS_ToUint32(ctx, &nbSubscriber, argv[1]))
        return JS_EXCEPTION;

    currentSendMessage = UA_STRING("");
    currentSendLog = UA_STRING("");
    droneObjectIdList = (JSValue *) malloc((nbDrone + nbSubscriber) * sizeof(JSValueConst));
    return JS_NewInt32(ctx, 0);
}

static JSValue js_stop_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    pubsubShouldRun = false;
    while(!pubsubExited)
        sleep(1);
    free(droneObjectIdList);

    cleanQueue(&sendMessageQueue);
    clear_str(&currentSendMessage);
    cleanQueue(&sendLogQueue);
    clear_str(&currentSendLog);

    return JS_NewInt32(ctx, 0);
}

// Connexion management functions
#ifdef WITH_AUTOPILOT
static JSValue js_start(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    const char *ip;
    const char *mavsdk_log_file;
    int port;
    int timeout;
    int res;

    ip = JS_ToCString(ctx, argv[0]);
    if (JS_ToInt32(ctx, &port, argv[1]))
        return JS_EXCEPTION;
    mavsdk_log_file = JS_ToCString(ctx, argv[2]);
    if (JS_ToInt32(ctx, &timeout, argv[3]))
        return JS_EXCEPTION;

    res = start(ip, port, mavsdk_log_file, timeout);
    JS_FreeCString(ctx, ip);
    JS_FreeCString(ctx, mavsdk_log_file);

    return JS_NewInt32(ctx, res);
}

static JSValue js_stop(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, stop());
}

static JSValue js_reboot(JSContext *ctx, JSValueConst thisVal,
                         int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, reboot());
}

// Flight state management functions

static JSValue js_arm(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, arm());
}

static JSValue js_takeOff(JSContext *ctx, JSValueConst thisVal,
                          int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, takeOff());
}

static JSValue js_land(JSContext *ctx, JSValueConst thisVal,
                       int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, land());
}

// Flight management functions

static JSValue js_loiter(JSContext *ctx, JSValueConst thisVal,
                         int argc, JSValueConst *argv)
{
    double latitude;
    double longitude;
    double altitude;
    double radius;
    double speed;

    if (JS_ToFloat64(ctx, &latitude, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &longitude, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &altitude, argv[2]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &radius, argv[3]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &speed, argv[4]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, loiter(latitude, longitude, (float)altitude, (float)radius, (float)speed));
}

static JSValue js_setTargetCoordinates(JSContext *ctx,
                                       JSValueConst thisVal,
                                       int argc, JSValueConst *argv)
{
    double latitude;
    double longitude;
    double altitude;
    double speed;

    if (JS_ToFloat64(ctx, &latitude, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &longitude, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &altitude, argv[2]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &speed, argv[3]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetCoordinates(latitude, longitude, (float)altitude, (float)speed));
}

// Information functions

static JSValue js_getAltitude(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getAltitude());
}

static JSValue js_getInitialAltitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialAltitude());
}

static JSValue js_getInitialLatitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLatitude());
}

static JSValue js_getInitialLongitude(JSContext *ctx,
                                      JSValueConst thisVal,
                                      int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLongitude());
}

static JSValue js_getTakeOffAltitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getTakeOffAltitude());
}

static JSValue js_getYaw(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getYaw());
}

static JSValue js_getSpeed(JSContext *ctx, JSValueConst thisVal,
                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getSpeed());
}

static JSValue js_getClimbRate(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getClimbRate());
}

static JSValue js_gpsIsOk(JSContext *ctx, JSValueConst this_val,
                          int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, gpsIsOk());
}

static JSValue js_isReadyToFly(JSContext *ctx, JSValueConst this_val,
                               int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, isReadyToFly());
}

static JSValue js_isLanding(JSContext *ctx, JSValueConst this_val,
                            int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, isLanding());
}

static JSValue js_updateLogAndProjection(JSContext *ctx, JSValueConst this_val,
                                         int argc, JSValueConst *argv)
{
    updateLogAndProjection();
    return JS_NewInt32(ctx, 0);
}
#endif

static const JSCFunctionListEntry js_funcs[] = {
    JS_CFUNC_DEF("setMessage", 1, js_drone_set_message ),
    JS_CFUNC_DEF("setLog", 1, js_drone_set_log ),
    JS_CFUNC_DEF("runPubsub", 6, js_run_pubsub ),
    JS_CFUNC_DEF("stopPubsub", 0, js_stop_pubsub ),
    JS_CFUNC_DEF("initPubsub", 2, js_init_pubsub ),
#ifdef WITH_AUTOPILOT
    JS_CFUNC_DEF("start", 4, js_start ),
    JS_CFUNC_DEF("stop", 0, js_stop ),
    JS_CFUNC_DEF("reboot", 0, js_reboot ),
    JS_CFUNC_DEF("arm", 0, js_arm ),
    JS_CFUNC_DEF("takeOff", 0, js_takeOff ),
    JS_CFUNC_DEF("land", 0, js_land ),
    JS_CFUNC_DEF("loiter", 5, js_loiter ),
    JS_CFUNC_DEF("setTargetCoordinates", 4, js_setTargetCoordinates ),
    JS_CFUNC_DEF("getPosition", 0, js_new_position ),
    JS_CFUNC_DEF("getAltitude", 0, js_getAltitude ),
    JS_CFUNC_DEF("getInitialAltitude", 0, js_getInitialAltitude ),
    JS_CFUNC_DEF("getInitialLatitude", 0, js_getInitialLatitude ),
    JS_CFUNC_DEF("getInitialLongitude", 0, js_getInitialLongitude ),
    JS_CFUNC_DEF("getTakeOffAltitude", 0, js_getTakeOffAltitude ),
    JS_CFUNC_DEF("getYaw", 0, js_getYaw ),
    JS_CFUNC_DEF("getSpeed", 0, js_getSpeed ),
    JS_CFUNC_DEF("getClimbRate", 0, js_getClimbRate ),
    JS_CFUNC_DEF("gpsIsOk", 0, js_gpsIsOk ),
    JS_CFUNC_DEF("isReadyToFly", 0, js_isReadyToFly ),
    JS_CFUNC_DEF("isLanding", 0, js_isLanding ),
    JS_CFUNC_DEF("updateLogAndProjection", 0, js_updateLogAndProjection ),
#endif
};

static int js_init(JSContext *ctx, JSModuleDef *m)
{
    JSValue droneProto, droneClass, positionProto;

    JS_NewClassID(&jsDroneClassId);
    JS_NewClassID(&jsPositionClassId);
    JS_NewClass(JS_GetRuntime(ctx), jsDroneClassId, &jsDroneClass);
    JS_NewClass(JS_GetRuntime(ctx), jsPositionClassId, &jsPositionClass);

    droneProto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, droneProto, js_drone_proto_funcs,
                               countof(js_drone_proto_funcs));
    positionProto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, positionProto, js_position_proto_funcs,
                               countof(js_position_proto_funcs));

    droneClass = JS_NewCFunction2(ctx, js_drone_ctor, "Drone", 1,
                                   JS_CFUNC_constructor, 0);
    JS_SetConstructor(ctx, droneClass, droneProto);
    JS_SetClassProto(ctx, jsDroneClassId, droneProto);
    JS_SetClassProto(ctx, jsPositionClassId, positionProto);

    JS_SetModuleExport(ctx, m, "Drone", droneClass);

    return JS_SetModuleExportList(ctx, m, js_funcs,
                                  countof(js_funcs));
}

JSModuleDef *js_init_module(JSContext *ctx, const char *module_name)
{
    JSModuleDef *m;
    m = JS_NewCModule(ctx, module_name, js_init);
    if (!m)
        return NULL;
    JS_AddModuleExportList(ctx, m, js_funcs, countof(js_funcs));
    JS_AddModuleExport(ctx, m, "Drone");
    return m;
}
